/*
 * bash-prompt.c
 * -------------
 * A simple C program to output a functional line above the PS1 prompt,
 * with the hostname and current working path on the left and the
 * date/time (formatted for your locale) on the right of the screen.
 * If the path becomes too long, it will truncate it and add "..."
 * 
 * It also sends a standard terminal control sequence that sets the
 * terminal title to the hostname (very convenient if you have a lot
 * of shells open in different tabs of your terminal program.)
 *
 * Copyright (C)2014 Daniel G. Delaney
 *
 * Written by:     Daniel G. Delaney (fluidmind.org)
 * Version:        2.0.1
 * Last updated:   Jan, 2014
 * Status:         Freeware
 *                 This program may be freely used and distributed
 *                 as long as this copyright information is not modified.
 *                 If you make improvements, please contact the author
 *                 and suggest them for the official distribution.
 *
 *
 * COMPILING:
 * Nothing special about compiling this little program:
 *
 *    cc bash-prompt.c -o bash-prompt
 *
 */


#include <stdio.h>
#include <string.h>
#include <wchar.h>      // for handling UTF-8 directory names
#include <sys/ioctl.h>  // for getting the terminal width
#include <stdlib.h>	    // for getenv()
#include <unistd.h>	    // for getcwd()
#include <time.h>
#include <locale.h>


int main (void)
{
    setlocale(LC_ALL, "");

    const size_t    WCHAR_SIZE = sizeof(wchar_t);
    struct winsize  w;
    size_t          screen_width;
    size_t          cwd_space;
    time_t          current_timestamp;
    char            time_string[25];
    struct tm       *local_time;
    size_t          time_length;
    char            *home_dir = getenv("HOME");
    char            hostname[64];
    char            term_esc_sequence[70];
    char            *cwd;
    size_t          cwd_length;
    char            cwd_prefix[67];
    size_t          cwd_prefix_length;
    size_t          total_cwd_length;
    wchar_t         *w_cwd;
    wchar_t         w_cwd_prefix[67];
    char            *temp;
    wchar_t         *w_temp;



    // Get the column width of the screen
    ioctl(0, TIOCGWINSZ, &w);
    screen_width = w.ws_col;

    // Determine how much space we have for hostname:cwd
    cwd_space = screen_width - 27;

    // Get the date and time, formatted for the current locale
    current_timestamp = time(NULL);
    local_time = localtime(&current_timestamp);
    strftime(time_string, sizeof(time_string), "%x %X", local_time);
    time_length = strlen(time_string);

    // Add the hostname to the beginning of the current working directory
    gethostname(hostname, 63);
    strncpy(cwd_prefix, hostname, 63);
    strcat(cwd_prefix, ":");

    // Get the current working directory
    cwd = getcwd(NULL, 0);

    // If the cwd has been removed by another process, we have a problem
    if (cwd == NULL) {
        cwd = malloc(36);
        strcpy(cwd, "[current working directory unknown]");
    } else {
        // Add a slash at the end, just because I like it that way :-)
        if (strcmp(cwd, "/") != 0) {
            cwd = realloc(cwd, strlen(cwd) + 2);
            strcat(cwd, "/"); 
            cwd_length = strlen(cwd);
        }

        // If cwd starts with the user's home directory, replace it with '~'
        if (strlen(cwd) > strlen(home_dir)) {
            temp = strndup(cwd, strlen(home_dir));
            if (strcmp(temp, home_dir) == 0) {
                strncat(cwd_prefix, "~/", 66 - strlen(cwd_prefix));
                memmove(cwd, &cwd[strlen(home_dir) + 1], cwd_length - strlen(home_dir));
                cwd_length = strlen(cwd);
            }
            free(temp);
        }
    }

    // Convert the cwd_prefix to wide-character codes and get the length
    cwd_prefix_length = mbstowcs(w_cwd_prefix, cwd_prefix, 66);

    // Convert the cwd to wide-character codes and get the length
    cwd_length = mbstowcs(NULL, cwd, 0);
    w_cwd = (wchar_t *)malloc((cwd_length + 1) * WCHAR_SIZE);
    mbstowcs(w_cwd, cwd, cwd_length + 1);

    // Get the total length of what will be displayed
    total_cwd_length = cwd_prefix_length + cwd_length;

    // Is the cwd too long for the space that's available for it?
    if (total_cwd_length > cwd_space) {
        // Truncate the left side of it, but leave space for the ellipses
        wmemmove(w_cwd + 3, &w_cwd[(total_cwd_length - cwd_space) + 3],
                            (cwd_length - (total_cwd_length - cwd_space) - 2) );

        // Then add an ellipses to the beginning of it
        wmemcpy(w_cwd, L"...", 3);

        cwd_length = wcslen(w_cwd);   // this will have changed
    }

    // Set the escape sequence to put the hostname in the title of the terminal
    strcpy(term_esc_sequence, "\033]0;");
    strncat(term_esc_sequence, hostname, 63);
    strcat(term_esc_sequence, "\007");

    w_temp = (wchar_t *)malloc((cwd_prefix_length + cwd_length + 1) * WCHAR_SIZE);
    wcscpy(w_temp, w_cwd_prefix);
    wcscat(w_temp, w_cwd);

    // Lastly, adjust cwd_space, because printf thinks it's for single-byte characters
    cwd_space += (wcstombs(NULL, w_temp, 0) - total_cwd_length);


    // Just the first line (set the second line in PS1)
    printf("%s\n%-*ls%26s\n", term_esc_sequence, (unsigned)cwd_space, w_temp, time_string);


    // Clean up
    free(w_temp);
    free(cwd);
    free(w_cwd);

    return 1;
}
